"""Main class for sleep tools managment."""

import numpy as np
from PyQt5 import QtWidgets
from sleep_scoring.utils import (rereferencing, bipolarization, find_non_eeg,
                            commonaverage)


class UiTools(object):
    """Main class for sleep tools managment."""

    def __init__(self):
        """Init."""
        self._tool_pick.currentIndexChanged.connect(self._fcn_tool_pick)
        # Find non-eeg channels :
        self._noneeg = find_non_eeg(self._channels)
        # =====================================================================
        # RE-REFERENCING
        # =====================================================================
        # Add channels to scrolling area :
        self._ToolsRefIgnArea.setVisible(False)
        self._reChecks = []
        for i, k in enumerate(self._channels):
            if not self._noneeg[i]:
                # Add a checkbox to the scrolling panel :
                box = QtWidgets.QCheckBox(self._PanScrollChan)
                # Name checkbox with channel name :
                box.setText(k)
                # Get it :
                self._reChecks.append(box)
                # Add checkbox to the grid :
                self._ToolsRefIgnGrd.addWidget(box, i, 0, 1, 1)
        # Connections :
        self._ToolsRefIgn.clicked.connect(self._fcn_ref_chan_ignore)
        self._ToolsRefLst.addItems(np.array(self._channels)[~self._noneeg])
        self._ToolsRefMeth.currentIndexChanged.connect(self._fcn_ref_switch)
        self._ToolsRefApply.clicked.connect(self._fcn_ref_apply)
        self._fcn_ref_switch()

        # =====================================================================
        # MEAN / TREND
        # =====================================================================
        self._SigMean.clicked.connect(self._fcn_sig_processing)
        self._SigTrend.clicked.connect(self._fcn_sig_processing)

        # =====================================================================
        # FILTERING
        # =====================================================================
        self._SigFiltChan.addItems(self._channels)
        self._SigFiltDisp.currentIndexChanged.connect(self._fcn_filt_disp_as)
        self._SigFiltApply.clicked.connect(self._fcn_sig_processing)
        self._SigFilt.clicked.connect(self._fcn_filt_viz)
        self._SigFiltBand.currentIndexChanged.connect(self._fcn_filt_band)

    def _fcn_tool_pick(self):
        """Change tool type."""
        idx = int(self._tool_pick.currentIndex())
        self._stacked_tools.setCurrentIndex(idx)

    # =====================================================================
    # RE-REFERENCING
    # =====================================================================
    def _fcn_ref_switch(self):
        """Switch between re-referencing methods."""
        idx = int(self._ToolsRefMeth.currentIndex())
        # Single channel :
        if idx == 0:  # Single channel
            self._ToolsRefSingleW.setVisible(True)
        elif idx == 1:  # Common average
            self._ToolsRefSingleW.setVisible(False)
        elif idx == 2:  # Bipolarization
            self._ToolsRefSingleW.setVisible(False)

    def _fcn_ref_chan_ignore(self):
        """Display / hide list of channels to ignore."""
        self._ToolsRefIgnArea.setVisible(self._ToolsRefIgn.isChecked())

    def _fcn_ref_apply(self):
        """Apply re-referencing."""
        # By default, ingore non-eeg channel :
        to_ignore = self._noneeg
        if self._ToolsRefIgn.isChecked():
            for num, k in enumerate(self._reChecks):
                # Get the position of this channel :
                idinlst = self._channels.index(str(k.text()))
                # Set to ignore :
                to_ignore[idinlst] = k.isChecked()

        # Get the current selected method :
        idx = int(self._ToolsRefMeth.currentIndex())
        # Single channel :
        if idx == 0:  # Single channel
            # Get selected channel :
            idchan = idx = self._ToolsRefLst.currentIndex()
            # Re-referencing :
            self._data, self._channels, consider = rereferencing(
                self._data, self._channels, idchan,
                to_ignore)
            self._chanChecks[idx].setChecked(False)
        elif idx == 1:  # Common average
            self._data, self._channels, consider = commonaverage(
                self._data, self._channels, to_ignore)
        elif idx == 2:  # Bipolarization
            self._data, self._channels, consider = bipolarization(
                self._data, self._channels,
                to_ignore)

        # ____________________ Update ____________________
        a_max = np.argmax(consider)
        # Update data info :
        self._get_data_info()

        # Update and clear detections :
        self._DetectLocations.setRowCount(0)
        self._DetectChanSw.clear()
        self._detect.update_keys(self._channels)
        self._detect.reset()

        # Disconnect and clear listbox :
        self._PanSpecChan.currentIndexChanged.disconnect()
        self._PanSpecChan.clear()
        self._ToolDetectChan.clear()
        self._PanSpecChan.addItems(self._channels)
        self._ToolDetectChan.addItems(self._channels)
        # Reconnect :
        self._PanSpecChan.setCurrentIndex(a_max)
        self._ToolDetectChan.setCurrentIndex(a_max)

        # Update channel names :
        for num, k in enumerate(self._channels):
            self._chanChecks[num].setText(k)
            self._chanLabels[num].setText(k)

        # Ignore non re-referenced channels :
        if self._ToolsRefIgnore.isChecked():
            for num, k in enumerate(consider):
                # Remove from visible channels :
                self._chanChecks[num].setChecked(False)
                self._chanChecks[num].setVisible(k)
                self._chanLabels[num].setVisible(k)
                self._yminSpin[num].setVisible(k)
                self._ymaxSpin[num].setVisible(k)
                self._amplitudeTxt[num].setVisible(k)
                # Remove from chan list :
                self._PanSpecChan.model().item(num).setEnabled(k)
                self._ToolDetectChan.model().item(num).setEnabled(k)
        if not any([k.isChecked() for k in self._chanChecks]):
            self._chanChecks[a_max].setChecked(True)
        # Reconnect :
        self._PanSpecChan.currentIndexChanged.connect(self._fcn_spec_set_data)

        self._chan.update()
        self._fcn_chan_viz()

        # Finally disable GroupBox :
        self._ToolsRefGrp.setEnabled(False)

    # =====================================================================
    # DEMEAN / DETREND / FILTERING
    # =====================================================================
    def _fcn_sig_processing(self):
        """Signal processing function."""
        # ========== VALUES ==========
        # Mean and trend :
        demean = self._SigMean.isChecked()
        detrend = self._SigTrend.isChecked()
        # Filtering :
        dispas = self._SigFiltDisp.currentText()
        filt = self._SigFilt.isChecked()
        fstart = self._SigFiltFrom.value()
        fend = self._SigFiltTo.value()
        filttype = str(self._SigFiltMeth.currentText())
        filtorder = self._SigFiltOrder.value()
        filtband = str(self._SigFiltBand.currentText())

        # Filt a specific channel :
        channel = int(self._SigFiltChan.currentIndex()) - 1
        if channel >= 0:
            self._chanChecks[channel].setChecked(True)
            self._fcn_chan_viz()

        # ========== CHANNELS ==========
        # ---- Demean / detrend ----
        self._chan._preproc_channel = channel
        self._chan.demean = demean
        self._chan.detrend = detrend
        # ---- Filtering ----
        self._chan.dispas = dispas
        self._chan.filt = filt
        self._chan.fstart = fstart
        self._chan.fend = fend
        self._chan.forder = filtorder
        self._chan.filt_type = filttype
        self._chan.filt_band = filtband

        self._chan.update()

        # ========== SPECTROGRAM ==========
        chan = self._PanSpecChan.currentIndex()
        # ---- Demean / detrend ----
        self._spec.demean = demean
        self._spec.detrend = detrend
        # ---- Filtering ----
        self._spec.dispas = dispas
        self._spec.filt = filt and channel in [-1, chan]
        self._spec.fstart = fstart
        self._spec.fend = fend
        self._spec.forder = filtorder
        self._spec.filt_type = filttype
        self._spec.filt_band = filtband

        self._spec.update()

    def _fcn_filt_disp_as(self):
        """Display / hide settings for amplitude / phase / power."""
        # Get current type :
        dispas = self._SigFiltDisp.currentText()
        if dispas == 'filter':
            self._SigFiltBand.setEnabled(True)
            self._SigFiltMeth.setEnabled(True)
            self._SigFiltOrder.setEnabled(True)
            self._fcn_filt_band()
        else:
            self._SigFiltBand.setEnabled(False)
            self._SigFiltMeth.setEnabled(False)
            self._SigFiltOrder.setEnabled(False)
        # Run filter :
        self._fcn_sig_processing()

    def _fcn_filt_viz(self):
        """Display / hide filtering panel."""
        viz = self._SigFilt.isChecked()
        self._SigFiltW.setEnabled(viz)
        if not viz:
            self._fcn_sig_processing()

    def _fcn_filt_band(self):
        """Configure visible [fstart, fend] for the band possibilities."""
        # Get selected band :
        filtband = str(self._SigFiltBand.currentText())
        # Enable / disable [fstart, fend] :
        if filtband in ['bandpass', 'bandstop']:
            fstart, fend = True, True
        elif filtband == 'lowpass':
            fstart, fend = False, True
        elif filtband == 'highpass':
            fstart, fend = True, False
        # Set enable fstart / fend :
        self._SigFiltFrom.setEnabled(fstart)
        self._SigFiltTo.setEnabled(fend)



"""Main class for sleep tools managment."""
import numpy as np
from PyQt5 import QtWidgets, QtCore
import logging

from sleep_scoring.utils import (remdetect, spindlesdetect, slowwavedetect,
                            kcdetect, peakdetect, mtdetect)
from sleep_scoring.utils.sleep.event import _events_to_index

logger = logging.getLogger('visbrain')


USER_METHOD = {'Spindles': 'spindle', 'Slow waves': 'sw', 'K-complexes': 'kc',
               'REM': 'rem', 'Muscle twitches': 'mt', 'Peaks': 'peak'}


class UiDetection(object):
    """Main class for sleep tools managment."""

    def __init__(self):
        """Init."""
        # =====================================================================
        # REM / SPINDLES / KC / SW / MT DETECTION
        # =====================================================================
        # Commonth elements :
        self._ToolDetectChan.addItems(self._channels)
        self._ToolDetectType.currentIndexChanged.connect(
            self._fcn_switch_detection)
        self._ToolDetectApply.clicked.connect(self._fcn_apply_detection)
        # Apply method (Selected / Visible / All) :
        self._ToolRdSelected.clicked.connect(self._fcn_apply_method)
        self._ToolRdViz.clicked.connect(self._fcn_apply_method)
        self._ToolRdAll.clicked.connect(self._fcn_apply_method)
        self._ToolDetectProgress.hide()
        self._fcn_switch_detection()

        # -------------------------------------------------
        # Location table :
        self._DetectChanSw.currentIndexChanged.connect(
            self._fcn_run_switch_location)
        self._DetectRm.clicked.connect(self._fcn_rm_location)
        self._DetectViz.clicked.connect(self._fcn_viz_location)
        self._DetecRmEvent.clicked.connect(self._fcn_rm_selected_event)
        self._DetectLocations.itemSelectionChanged.connect(
            self._fcn_goto_location)
        self._DetectLocations.cellChanged.connect(self._fcn_edit_detection)
        self._DetectionTab.setTabEnabled(1, False)

    # =====================================================================
    # ENABLE / DISABLE GUI COMPONENTS (based on selected channels)
    # =====================================================================
    def _fcn_apply_method(self):
        """Be sure to apply hypnogram report only on selected channel."""
        viz = self._ToolRdSelected.isChecked()
        self._ToolDetectChan.setEnabled(viz)

    # =====================================================================
    # SWITCH DETECTION TYPE
    # =====================================================================
    def _fcn_switch_detection(self):
        """Switch between detection types (show / hide panels)."""
        idx = int(self._ToolDetectType.currentIndex())
        method = str(self._ToolDetectType.currentText())
        self._stacked_detections.setCurrentIndex(idx)
        enable = USER_METHOD[method] not in self._custom_detections.keys()
        self._stacked_detections.setEnabled(enable)

    # =====================================================================
    # RUN DETECTION
    # =====================================================================
    # -------------- Get channels to apply detection --------------
    def _fcn_get_chan_detection(self):
        """Get on which channel to apply the detection."""
        # Selected channel :
        if self._ToolRdSelected.isChecked():
            idx = [self._ToolDetectChan.currentIndex()]

        # Visible channels :
        elif self._ToolRdViz.isChecked():
            idx = [
                k for k in range(len(self)) if self._chanWidget[k].isVisible()]

        # All channels :
        elif self._ToolRdAll.isChecked():
            idx = range(len(self))

        return idx

    # -------------- Get the function to run --------------
    def _fcn_get_detection_function(self, method):
        """Get the method to use for the detection (default or custom).

        Parameters
        ----------
        method : string
            Method to use.
        """
        user_method = USER_METHOD[method]
        if user_method in self._custom_detections.keys():
            logger.warning("Custom method used for %s detection" % method)
            fcn = self._custom_detections[user_method]
        else:
            logger.info("Default method used for %s detection" % method)
            # Switch between detection types :
            if method == 'REM':
                th = self._ToolRemTh.value()
                rem_only = self._ToolRemOnly.isChecked()
                def fcn(data, sf, time, hypno):  # noqa
                    return remdetect(data, sf, hypno, rem_only, th)
            elif method == 'Spindles':
                thr = self._ToolSpinTh.value()
                fmin = self._ToolSpinFmin.value()
                fmax = self._ToolSpinFmax.value()
                tmin = self._ToolSpinTmin.value()
                tmax = self._ToolSpinTmax.value()
                nrem_only = self._ToolSpinRemOnly.isChecked()
                def fcn(data, sf, time, hypno):  # noqa
                    return spindlesdetect(data, sf, thr, hypno, nrem_only,
                                          fmin, fmax, tmin, tmax)
            elif method == 'Slow waves':
                thr = self._ToolWaveTh.value()
                def fcn(data, sf, time, hypno):  # noqa
                    return slowwavedetect(data, sf, thr)
            elif method == 'K-complexes':
                proba_thr = self._ToolKCProbTh.value()
                amp_thr = self._ToolKCAmpTh.value()
                tmin = self._ToolKCMinDur.value()
                tmax = self._ToolKCMaxDur.value()
                min_amp = self._ToolKCMinAmp.value()
                max_amp = self._ToolKCMaxAmp.value()
                nrem_only = self._ToolKCNremOnly.isChecked()
                def fcn(data, sf, time, hypno):  # noqa
                    return kcdetect(data, sf, proba_thr, amp_thr, hypno,
                                    nrem_only, tmin, tmax, min_amp, max_amp)
            elif method == 'Muscle twitches':
                th = self._ToolMTTh.value()
                rem_only = self._ToolMTOnly.isChecked()
                def fcn(data, sf, time, hypno):  # noqa
                    return mtdetect(data, sf, th, hypno, rem_only)
            elif method == 'Peaks':
                look = int(self._ToolPeakLook.value() * self._sf)
                _disp = self._ToolPeakMinMax.currentIndex()
                disp = ['max', 'min', 'minmax'][_disp]
                def fcn(data, sf, time, hypno):  # noqa
                    return peakdetect(sf, data, self._time, look, 1., disp,
                                      'auto')

        def fcn_check(data, sf, time, hypno):
            """Wrap fcn with type checking."""
            assert isinstance(data, np.ndarray)
            idx = fcn(data, sf, time, hypno)
            idx = np.asarray(idx)
            if not idx.size:
                return idx
            # Check indices shape and format to (n_events, 2) :
            if (idx.ndim == 2) and (idx.shape[1] == 2):  # (n_events, 2)
                return idx.astype(int)
            elif idx.ndim == 1:  # 1d vector
                if idx.dtype == bool:  # boolean array
                    assert len(idx) == len(data)
                    idx = np.arange(len(data))[idx]
                return _events_to_index(idx)
            else:
                raise ValueError("Return indices should either be an (n_events"
                                 ", 2) array or a boolean array of shape "
                                 "(n_time_points,) or an array with "
                                 "consecutive detected events.")

        return fcn_check

    # -------------- Run detection (only on selected channels) --------------
    def _fcn_apply_detection(self):
        """Apply detection (either REM/Spindles/Peaks/SlowWave/KC/MT)."""
        # Get channels to apply detection and the detection method :
        idx = self._fcn_get_chan_detection()
        method = str(self._ToolDetectType.currentText())

        fcn = self._fcn_get_detection_function(method)

        ############################################################
        # RUN DETECTION
        ############################################################
        for i, k in enumerate(idx):
            # Display progress bar (only if needed):
            if len(idx) > 1:
                self._ToolDetectProgress.show()

            # Run detection :
            index = fcn(self._data[k, :], self._sf, self._time, self._hypno)
            nb = index.shape[0]
            dty = nb / (len(self._time) / self._sf / 60.)
            # dur = (index[:, 1] - index[:, 0]) * (1000. / self._sf)

            logger.info(("Perform %s detection on channel %s. %i events "
                         "detected.") % (method, self._channels[k], nb))

            if index.size:
                # Enable detection tab :
                self._DetectionTab.setTabEnabled(1, True)
                self._detect.dict[(self._channels[k], method)]['index'] = index
                # Be sure panel is displayed :
                if not self._canvas_is_visible(k):
                    self._canvas_set_visible(k, True)
                    self._chan.visible[k] = True
                self._chan.loc[k].visible = True
                # Update plot :
                self._fcn_slider_move()

            # Update progress bar :
            # TODO: casted to builtin int
            self._ToolDetectProgress.setValue(int(100. * (i + 1) / len(self)))

        ############################################################
        # NUMBER // DENSITY
        ############################################################
        if index.size:
            # Report results on table :
            self._ToolDetectTable.setRowCount(1)
            self._ToolDetectTable.setItem(0, 0, QtWidgets.QTableWidgetItem(
                str(nb)))
            self._ToolDetectTable.setItem(0, 1, QtWidgets.QTableWidgetItem(
                str(round(dty, 2))))
        else:
            logger.error("No %s detected on channel %s. Adjust "
                         "parameters." % (method, self._channels[k]))

        ############################################################
        # LINE REPORT :
        ############################################################
        self._loc_line_report(select=True)

        # Activate the save detections menu and activate detection tab :
        self._check_detect_menu()

        # Finally, hide progress bar :
        self._ToolDetectProgress.hide()

    def _loc_line_report(self, *args, refresh=True, select=False):
        """Update line report."""
        self._detect.build_line(self._data)
        chans = self._detect.nonzero()
        if refresh:
            # Disconnect the table :
            self._DetectChanSw.disconnect()
            # Clear already existing events :
            self._DetectChanSw.clear()
            # Add list of detected event as (chan - type) :
            lst = []
            for k in chans.keys():
                for i in chans[k]:
                    lst.append(k + ' - ' + i)
            self._DetectChanSw.addItems(lst)
        self._fcn_run_switch_location()
        # Disable menu and detection tab if there is no detection :
        if self._DetectChanSw.count() == 0:
            self._DetectionTab.setTabEnabled(1, False)
        else:
            self._DetectionTab.setTabEnabled(1, True)
        self._check_detect_menu()
        # Reconnect table :
        self._DetectChanSw.currentIndexChanged.connect(
            self._fcn_run_switch_location)
        # Select the last detected event :
        if refresh and select:
            self._DetectChanSw.setCurrentIndex(self._DetectChanSw.count() - 1)

    # =====================================================================
    # FILL LOCATION TABLE
    # =====================================================================
    def _get_current_chan_type(self):
        """Return the currently selected channel and detection type."""
        chtp = tuple(str(self._DetectChanSw.currentText()).rsplit(' - '))
        return chtp if len(chtp) == 2 else ('', '')

    def __get_visible_loc(self):
        """Enable/Disable detection table accroding to the visibility."""
        # Get the currently selected channel and type :
        chan, tps = self._get_current_chan_type()
        if chan and tps:
            if tps == 'Peaks':
                viz = self._detect.peaks[(chan, tps)].visible
            else:
                viz = self._detect.line[(chan, tps)].visible
            self._DetectViz.setChecked(viz)
            self._DetectLocations.setEnabled(viz)

    def _fcn_rm_location(self):
        """Remove a detection."""
        # Get the currently selected channel and type :
        chan, types = self._get_current_chan_type()
        # Remove detection :
        if chan and types:
            self._detect.delete(chan, types)
            # Remove vertical indicators :
            pos = np.full((1, 3), -10., dtype=np.float32)
            self._chan.loc[self._channels.index(chan)].set_data(pos=pos)
            # Clean table :
            self._DetectLocations.setRowCount(0)
            # Update GUI :
            self._loc_line_report()
        else:
            self._DetectionTab.setTabEnabled(1, False)
            self._check_detect_menu()

    def _fcn_viz_location(self):
        """Display/hide detections."""
        # Get the currently selected channel and type :
        chan, types = self._get_current_chan_type()
        # Set visibility :
        if chan and types:
            viz = self._DetectViz.isChecked()
            self._detect.visible(viz, chan, types)
            self._chan.loc[self._channels.index(chan)].visible = viz
            self._DetectLocations.setEnabled(viz)

    def _fcn_run_switch_location(self):
        """Run switch location channel and type."""
        # Get the currently selected channel and type :
        chan, types = self._get_current_chan_type()
        if chan and types:
            # Enable/disable the location table :
            self.__get_visible_loc()
            # Find index and durations :
            index = self._detect[(chan, types)]['index']
            # Get durations :
            dur = (index[:, 1] - index[:, 0]) * (1000. / self._sf)
            # Set hypnogram data :
            self._detect.build_hyp(chan, types)
            # Fill location table :
            self._fcn_fill_locations(chan, types, index, dur)

    def _fcn_fill_locations(self, channel, kind, index, duration):
        """Fill the location table."""
        # Disconnect location table :
        self._DetectLocations.disconnect()
        ref = ['Wake', 'N1', 'N2', 'N3', 'REM', 'ART']
        # Clean table :
        self._DetectLocations.setRowCount(0)
        # Get starting index:
        sta_ind, end_ind = index[:, 0], index[:, 1]
        # Define the length of the table:
        self._DetectLocations.setRowCount(min(len(sta_ind), len(duration)))
        # Fill the table :
        for num, (k, j, i) in enumerate(zip(sta_ind, end_ind, duration)):
            # Starting :
            self._DetectLocations.setItem(num, 0, QtWidgets.QTableWidgetItem(
                str(self._time[k])))
            # Ending :
            self._DetectLocations.setItem(num, 1, QtWidgets.QTableWidgetItem(
                str(self._time[j])))
            # Duration :
            self._DetectLocations.setItem(num, 2, QtWidgets.QTableWidgetItem(
                str(i)))
            # Type :
            item = QtWidgets.QTableWidgetItem(ref[int(self._hypno[k])])
            item.setFlags(QtCore.Qt.ItemIsEnabled)
            self._DetectLocations.setItem(num, 3, item)
        # Go to the first detected event :
        self._DetectLocations.selectRow(0)
        self._fcn_goto_location()
        # Reconnect :
        self._DetectLocations.itemSelectionChanged.connect(
            self._fcn_goto_location)
        self._DetectLocations.cellChanged.connect(self._fcn_edit_detection)

    # =====================================================================
    # GO TO THE LOCATION
    # =====================================================================
    def _fcn_goto_location(self):
        """Go to the selected row REM / spindles / peak."""
        # Get the currently selected channel and type :
        chan, types = self._get_current_chan_type()
        # Get selected row and channel :
        row = self._DetectLocations.currentRow()
        ix = self._channels.index(chan)
        if row >= 0:
            # Get starting and ending point :
            sta = float(str(self._DetectLocations.item(row, 0).text()))
            end = float(str(self._DetectLocations.item(row, 1).text()))
            # Go to :
            self._SlGoto.setValue(sta)
            # Set vertical lines to the location :
            self._chan.set_location(self._sf, self._data[ix, :], ix, sta, end)

    def _fcn_edit_detection(self):
        """Executed function when the item is edited."""
        # Get the currently selected channel and type :
        chan, types = self._get_current_chan_type()
        # Get selected row and col :
        row = self._DetectLocations.currentRow()
        col = self._DetectLocations.currentColumn()
        val = self._DetectLocations.item(row, col).text()
        if col in [0, 1]:  # Edit starting/ending point
            val = int(np.round(float(val) * self._sf))
            self._detect[(chan, types)]['index'][row, col] = val
        elif col == 2:  # Edit duration
            val = int(np.round(float(val) * self._sf / 1000.))
            start = self._detect[(chan, types)]['index'][row, 0]
            self._detect[(chan, types)]['index'][row, 1] = start + val
        elif col == 3:  # Avoid stage editing
            self._DetectLocations
            self._DetectLocations.setItem(row, 3,
                                          QtWidgets.QTableWidgetItem(val))
        # Update :
        self._loc_line_report(refresh=False)
        self._DetectLocations.selectRow(row)

    def _fcn_rm_selected_event(self):
        """Remove the selected event in the table and update detections."""
        # Get selected row :
        row = self._DetectLocations.currentRow()  # return -1 when no more row
        if row + 1:
            # Remove row :
            self._DetectLocations.removeRow(row)
            # Get the currently selected channel and type :
            chan, types = self._get_current_chan_type()
            # Delete the selected event :
            index = self._detect[(chan, types)]['index']
            if not index.shape[0] - 1:
                self._detect.delete(chan, types)
                # Update :
                self._loc_line_report(refresh=True)
            else:
                self._detect[(chan, types)]['index'] = np.delete(index, row, 0)
                self._loc_line_report(refresh=False)
                self._DetectLocations.selectRow(row)



"""This script group the diffrent graphical components.

Grouped components :
    * PyQt elements (window, Pyqt functions...)
    * Vispy canvas functions
    * User shortcuts
"""
import numpy as np

from PyQt5 import QtWidgets
from vispy import app, scene
import vispy.visuals.transforms as vist

from ..visuals.marker import Markers
from .gui import Ui_MainWindow
from sleep_scoring.utils import color2vb


class UiInit(QtWidgets.QMainWindow, Ui_MainWindow, app.Canvas):
    """Group and initialize the graphical elements and interactions."""

    def __init__(self):
        """Init."""
        # Create the main window :
        super(UiInit, self).__init__(None)
        self.setupUi(self)


class TimeAxis(object):
    """Create a unique time axis."""

    def __init__(self, axis=True, x_label='', name='', bgcolor=(1., 1., 1.),
                 indic_color='darkred', fcn=[]):
        """Init."""
        # Create the main canvas :
        self.canvas = scene.SceneCanvas(keys=None, bgcolor=bgcolor,
                                        show=False)
        _ = [self.canvas.connect(k) for k in fcn]  # noqa

        # Create a grid :
        grid = self.canvas.central_widget.add_grid(margin=0)
        grid.spacing = 0

        # Add x-axis :
        self.xaxis = scene.AxisWidget(orientation='bottom', text_color='black')
        if axis:
            pad = grid.add_widget(row=0, col=0)
            pad.width_max = 50
        _col = int(axis)
        grid.add_widget(self.xaxis, row=1, col=_col)

        # Main plot :
        self.wc = grid.add_view(row=0, col=_col, border_color='black')

        # Add a square indicator :
        image = color2vb(indic_color)[np.newaxis, ...]
        self.mesh = scene.visuals.Image(image, name='Time indicator')
        self.mesh.transform = vist.STTransform()
        self.mesh.parent = self.wc.scene

        # Add markers :
        pos = np.full((1, 3), -10, dtype=np.float32)
        self.markers = Markers(pos=pos, parent=self.wc.scene)
        self.markers.set_gl_state('translucent')

    def set_data(self, tox=None, width=None, time=None, unit='seconds',
                 markers=None):
        """Move the main square."""
        # Get factor according to unit :
        if unit == 'seconds':
            fact = 1.
        elif unit == 'minutes':
            fact = 60.
        elif unit == 'hours':
            fact = 3660.
        # Move the square
        if (tox, width, time) != (None, None, None):
            self.mesh.transform.translate = tox / fact
            self.mesh.transform.scale = width / fact
            # Update camera :
            self.wc.camera.rect = (0, 0, (time.max() - time.min()) / fact, 1)
        # Set markers :
        if markers is not None:
            if markers.size:
                pos = np.zeros((len(markers), 3), dtype=np.float32)
                pos[:, 0] = markers / fact
                pos[:, 1] = .5
                pos[:, 2] = -10
            else:
                pos = np.full((1, 3), -10, dtype=np.float32)
            self.markers.set_data(pos=pos, symbol='triangle_down', size=20.,
                                  face_color='#42ab46', edge_width=0.)

    def set_camera(self, camera):
        """Set a camera and link all objects inside."""
        # Set camera to the main widget :
        self.wc.camera = camera
        # Link transformations with axis :
        self.xaxis.link_view(self.wc)

    def clean(self):
        """Clean axis."""
        self.mesh.parent = None
        self.mesh = None


class AxisCanvas(object):
    """Create a canvas with an embeded axis."""

    def __init__(self, axis=True, x_label='', name='', use_pad=False,
                 bgcolor='white', fcn=[]):
        """Init."""
        # Save variables :
        self._axis = axis

        # Create the main canvas :
        self.canvas = scene.SceneCanvas(keys=None, bgcolor=bgcolor, show=False,
                                        title=name)
        _ = [self.canvas.connect(k) for k in fcn]  # noqa

        # Add axis :
        if axis:
            # Create a grid :
            grid = self.canvas.central_widget.add_grid(margin=0)

            # Add y-axis :
            if use_pad:
                pad = grid.add_widget(row=1, col=0)
                pad.width_max = 50
            else:
                self.yaxis = scene.AxisWidget(orientation='left',
                                              text_color='black')
                self.yaxis.width_max = 50
                grid.add_widget(self.yaxis, row=1, col=0)

            # Add right padding :
            bottom_pad = grid.add_widget(row=2, col=0, row_span=1)
            bottom_pad.height_max = 10
            top_pad = grid.add_widget(row=0, col=0, row_span=1)
            top_pad.height_max = 10

            # Main plot :
            self.wc = grid.add_view(row=1, col=1, border_color='white')
        # Ignore axis :
        else:
            self.wc = self.canvas.central_widget.add_view()

    def set_camera(self, camera):
        """Set a camera and link all objects inside."""
        # Link transformations with axis :
        self.wc.camera = camera
        if hasattr(self, 'yaxis'):
            self.yaxis.link_view(self.wc)

    def visible_axis(self, visible=True):
        """Display or hide the axis."""
        self._axis = visible
        self.yaxis.visible = visible
        self._rpad.visible = visible



"""Save templates (brain, roi, volume...) to the tmp folder."""
import logging
import os
import numpy as np

from .path import path_to_visbrain_data, path_to_tmp
from ..utils.mesh import convert_meshdata

logger = logging.getLogger('visbrain')

__all__ = ['add_brain_template', 'remove_brain_template',
           'save_volume_template', 'remove_volume_template']


def add_brain_template(name, vertices, faces, normals=None, lr_index=None,
                       tmpfile=False):
    """Add a brain template to the default list.

    Parameters
    ----------
    name : string
        Name of the template.
    vertices : array_like
        Vertices of the template of shape (N, 3) or (N, 3, 3) if indexed by
        faces.
    faces : array_like
        Faces of the template of shape (M, 3)
    normals : array_like
        The normals of the template, with the same shape as vertices.
    lr_index : int | None
        Specify where to cut vertices for left and right hemisphere so that
        x_left <= lr_index and right > lr_index
    tmpfile : bool | False
        Specify if the saved brain template is a temporary file (in that case,
        saved in visbrain/data/tmp/templates).
    """
    # Convert meshdata :
    vertices, faces, normals = convert_meshdata(vertices, faces, normals)
    # Create templates folder if doesn't exist :
    vb_path = path_to_visbrain_data(folder='templates')
    if not os.path.exists(vb_path):
        os.mkdir(vb_path)
    # Get path to the templates/ folder :
    name = os.path.splitext(name)[0]
    if tmpfile:
        path = path_to_tmp(folder='templates', file=name + '.npz')
    else:
        path = path_to_visbrain_data(folder='templates', file=name + '.npz')
    # Save the template :
    np.savez_compressed(path, vertices=vertices, faces=faces, normals=normals,
                        lr_index=lr_index)
    logger.info("Brain template saved (%s)." % path)


def remove_brain_template(name):
    """Remove brain template from the default list.

    Parameters
    ----------
    name : string
        Name of the template to remove.
    """
    # Get path to the templates/ folder :
    name = os.path.splitext(name)[0]
    path = path_to_visbrain_data(folder='templates', file=name + '.npz')
    # Remove the file from templates/ folder :
    if os.path.isfile(path):
        os.remove(path)
        logger.info("Brain template removed (%s)." % path)
    else:
        raise ValueError("No file " + path)


def save_volume_template(name, vol, labels, index, hdr, tmpfile=False):
    """Save as a predefined ROI atlas.

    Parameters
    ----------
    name : string
        Name of the ROI atlas.
    vol : array_like
        The volume of shape (nx, ny, nz).
    labels : array_like
        Array of labels of type object and of length n_labels.
    index : array_like
        Array of index corresponding to the labels of type np.int and of length
        n_labels.
    hdr : array_like
        The matrix of transformation of shape (4, 4).
    """
    path_to_save = path_to_visbrain_data(folder='roi')
    if not os.path.exists(path_to_save):
        os.mkdir(path_to_save)
    if tmpfile:
        path_to_save = path_to_tmp(folder='roi', file=name + '.npz')
    else:
        path_to_save = path_to_visbrain_data(folder='roi', file=name + '.npz')
    np.savez_compressed(path_to_save, vol=vol, hdr=hdr, index=index,
                        labels=labels)
    logger.info("%s is now a default ROI object. Use `r_obj = RoiObj('%s')` to"
                " call it." % (name, name))


def remove_volume_template(name):
    """Remove a predefined ROI object.

    Parameters
    ----------
    name : string
        Name of the ROI atlas.
    """
    path_to_file = path_to_visbrain_data(folder='roi', file=name + '.npz')
    if os.path.isfile(path_to_file):
        os.remove(path_to_file)
        logger.info("%s ROI object removed." % name)



"""Test if dependencies are installed."""

__all__ = ('is_mne_installed', 'is_nibabel_installed', 'is_opengl_installed',
           'is_pandas_installed', 'is_lspopt_installed', 'is_xlrd_installed',
           'is_tensorpac_installed', 'is_sc_image_installed')


def _check_version(v_user, v_compare):
    """Compare package versions.

    Parameters
    ----------
    v_user : string
        Package version for the user.
    v_compare : string
        Package version to compare.
    """
    from distutils.version import LooseVersion
    return LooseVersion(v_user) >= LooseVersion(v_compare)


def is_mne_installed(raise_error=False):
    """Test if MNE is installed."""
    try:
        import mne  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("mne-python not installed. See https://github.com/mne-"
                      "tools/mne-python for installation instructions.")
    return is_installed


def is_nibabel_installed(raise_error=False):
    """Test if nibabel is installed."""
    try:
        import nibabel  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("nibabel not installed. See https://github.com/"
                      "nipy/nibabel for installation instructions.")
    return is_installed


def is_opengl_installed(raise_error=False):
    """Test if OpenGL is installed."""
    try:
        import OpenGL.GL as GL  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("OpenGL not installed. See http://vispy.org/installation"
                      ".html#backend-requirements for installation "
                      "instructions.")
    return is_installed


def is_pandas_installed(raise_error=False):
    """Test if pandas is installed."""
    try:
        import pandas  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("pandas not installed. See https://pandas.pydata.org/#"
                      "best-way-to-install for installation instructions.")
    return is_installed


def is_lspopt_installed(raise_error=False):
    """Test if lspopt is installed."""
    try:
        import lspopt  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("lspopt not installed. See https://github.com/hbldh/"
                      "lspopt#installation for installation instructions.")
    return is_installed


def is_tensorpac_installed(raise_error=False):
    """Test if tensorpac is installed."""
    try:
        import tensorpac  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("tensorpac not installed. See https://github.com/Etienne"
                      "Cmb/tensorpac#installation for installation "
                      "instructions.")
    return is_installed


def is_xlrd_installed(raise_error=False):
    """Test if xlrd is installed."""
    try:
        import xlrd  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("xlrd not installed. In a terminal, run : pip install "
                      "xlrd")
    return is_installed


def is_sc_image_installed(raise_error=False):
    """Test if scikit-image is installed."""
    try:
        import skimage  # noqa
        is_installed = True
    except:
        is_installed = False
    # Raise error (if needed) :
    if raise_error and not is_installed:
        raise IOError("scikit-image not installed. In a terminal, run : pip"
                      " install scikit-image")
    return is_installed



"""Export graphical components as figures.

- write_fig_hyp : Export the hypnogram as a figure
- write_fig_canvas : Export a canvas as a figure
- write_fig_pyqt : Export a GUI window as a figure
"""
import os
import logging
import numpy as np
from ..utils.color import color2vb

logger = logging.getLogger('visbrain')

__all__ = ('write_fig_hyp', 'write_fig_spindles', 'write_fig_canvas',
           'write_fig_pyqt')


def write_fig_hyp(data, sf, file=None, start_s=0, grid=False, ascolor=False,
                  dpi=300, colors={-1: '#8bbf56', 0: '#56bf8b', 1: '#aabcce',
                                   2: '#405c79', 3: '#0b1c2c', 4: '#bf5656'}):
    """Export hypnogram to a high-res png figure.

    Parameters
    ----------
    data : array_like
        Hypnogram vector
    sf : float
        The sampling frequency of displayed elements (could be the
        down-sampling frequency)
    file : string | None
        Output filename (with full path). If None, the plot is displayed.
    start_s : float | 0.
        Record starting time given in seconds.
    grid : bool | False
        Plot X and Y grid.
    ascolor : bool | False
        Plot in color
    dpi : int | 600
        Dots per inches
    color : dict | {}
        Color for each sleep stage. Default is : {-1: '#8bbf56', 0: '#56bf8b',
        1: '#aabcce', 2: '#405c79', 3: '#0b1c2c', 4: '#bf5656'}
    """
    import matplotlib.pyplot as plt
    import datetime

    # Internal copy :
    hypno = data.copy()

    # Downsample to get one value per second
    sf = int(sf)
    hypno = hypno[::sf]

    # Put REM between Wake and N1 sleep
    hypno[hypno >= 1] += 1
    hypno[hypno == 5] = 1
    idx_rem = np.where(hypno == 1)[0]
    val_rem = np.zeros(hypno.size)
    val_rem[:] = np.nan
    val_rem[idx_rem] = 1

    # Find if artefacts are present in hypno
    art = True if -1 in hypno else False

    # Start plotting
    fig, ax = plt.subplots(figsize=(8, 3), edgecolor='k')
    lhyp = len(hypno) / 60
    lw = 1.5
    if lhyp < 60:
        xticks = np.arange(0, len(hypno), 10 * 60)
        lw = 2
    elif lhyp < 180 and lhyp > 60:
        xticks = np.arange(0, len(hypno), 30 * 60)
    else:
        xticks = np.arange(0, len(hypno), 60 * 60)

    xticks = np.append(xticks, len(hypno))
    xlabels = (xticks + start_s).astype(int)
    xlabels_str = [str(datetime.timedelta(seconds=int(j)))[:-3]
                   for i, j in enumerate(xlabels)]
    xlabels_str = [s.replace('1 day, ', '') for s in xlabels_str]
    plt.xlim(0, len(hypno))
    plt.xticks(xticks, xlabels_str)
    if not ascolor:
        plt.plot(hypno, 'k', ls='steps', linewidth=lw)
    else:
        for k, i in colors.items():
            # Quick and dirty switch :
            if k == 1:
                q = 2
            elif k == 4:
                q = 1
            elif k in [2, 3]:
                q = k + 1
            else:
                q = k
            mask = np.ones((len(hypno),), dtype=bool)
            idxm = np.where(hypno == q)[0] + 1
            idxm[idxm >= len(hypno)] = len(hypno) - 1
            mask[idxm] = False
            # TODO: ls should be "drawstyle" instead
            # plt.plot(np.ma.masked_array(hypno, mask=mask), i, ls='steps',
            #          linewidth=lw)
            plt.plot(np.ma.masked_array(hypno, mask=mask), i, drawstyle='steps',
                     linewidth=lw)

    # Plot REM epochs
    remcol = 'k' if not ascolor else colors[4]
    for i in np.arange(0.6, 1, 0.01):
        plt.plot(np.arange(len(hypno)), i * val_rem, remcol, linewidth=lw)

    # Y-Ticks and Labels
    if art:
        ylabels = ['Art', 'Wake', 'REM', 'N1', 'N2', 'N3']
        plt.yticks([-1, 0, 1, 2, 3, 4], ylabels)
        plt.ylim(-1.5, 4.5)
    else:
        ylabels = ['', 'Wake', 'REM', 'N1', 'N2', 'N3']
        plt.yticks([-0.5, 0, 1, 2, 3, 4], ylabels)
        plt.ylim(-.5, 4.5)

    # X-Ticks and Labels
    plt.xlabel("Time")
    plt.ylabel("Sleep Stage")

    # Grid
    if grid:
        plt.grid(True, 'major', ls=':', lw=.2, c='k', alpha=.3)

    plt.tick_params(axis='both', which='both', bottom=True, top=False,
                    labelbottom=True, left=True, right=False, labelleft=True,
                    labelcolor='k', direction='out')

    # Invert Y axis and despine
    ax.invert_yaxis()
    ax.spines['right'].set_visible(False)
    ax.spines['top'].set_visible(False)
    ax.spines['left'].set_visible(True)
    ax.spines['bottom'].set_visible(True)

    ax.spines['left'].set_position(('outward', 10))
    ax.spines['bottom'].set_position(('outward', 10))
    # TODO: figure out how to set bounds
    # ax.spines['bottom'].set_smart_bounds(True)

    # Save as 600 dpi .png
    if isinstance(file, str):
        plt.savefig(file, format='png', dpi=dpi, bbox_inches='tight')
        logger.info('Image successfully saved (%s)' % file)
        plt.close()
    else:
        plt.show()


def write_fig_spindles(data, sf, hypno=None, file=None, start_s=0.,
                       window_s=10., thr=3., nrem_only=False, dpi=300):
    """Show steps of the spindles detection for a specific time window.

    Parameters
    ----------
    data : array_like
        Data vector
    sf : float
        The sampling frequency of displayed elements (could be the
        down-sampling frequency)
    hypno : array_like | None
        Hypnogram vector
    file : string | None
        Output filename (with full path). If None, the plot is displayed.
    start_s : float | 0.
        Starting point in sec of the window to plot
    window_s : float | 10.
        Window duration in seconds
    thresh : float | 3
        Hard threshold for the spindles detection
    dpi : int | 300
        Dots per inches
    """
    import matplotlib.pyplot as plt
    from ..utils.sleep import spindlesdetect
    from ..utils.filtering import filt

    # Define an empty hypno hypnogram if None :
    hypno = np.zeros_like(data) if hypno is None else hypno

    # Run spindles detection on the selected channel
    (idx_spindles, _, _, dur, pwr, idx_start, idx_stop, hard_thr, soft_thr,
     idx_sigma, fmin, fmax, sigma_nfpow, amplitude,
     sigma_thr) = spindlesdetect(data, sf, thr, hypno, nrem_only,
                                 return_full=True)

    # Define plotting range
    start_sf = int(start_s * sf)
    x = range(start_sf, start_sf + int(window_s * sf))

    # Bandpass filter of the window
    elec_filt = filt(sf, [12., 14.], data[x], order=4)

    # Find beginning and end of spindle within the window
    idx_start_win = idx_start[(idx_start >= min(x)) & (idx_start <= max(x))]
    idx_stop_win = idx_stop[(idx_stop >= min(x)) & (idx_stop <= max(x))]
    sp_in_win = np.in1d(idx_start, idx_start_win)
    sp_power = pwr[sp_in_win]
    sp_duration = dur[sp_in_win]

    # Find indices of spindles within the window
    idx_spindles_win = idx_spindles[
        (idx_spindles >= min(x)) & (idx_spindles <= max(x))]

    # Find indices of sigma power > supra-threshold within window
    idx_sigma_win = idx_sigma[(idx_sigma > min(x)) & (idx_sigma < max(x))]
    # Find indices of wavelet amplitude > supra-threshold within window
    with np.errstate(divide='ignore', invalid='ignore'):
        idx_hard = np.where(amplitude > hard_thr)[0]

    idx_hard_win = idx_hard[(idx_hard > min(x)) & (idx_hard < max(x))]

    # Initialize Y vector
    y_sigma, y_spindles, y_wlt, y_hard = np.empty(len(x)), np.empty(len(x)), \
        np.empty(len(x)), np.empty(len(x))
    y_sigma[:], y_spindles[:], y_wlt[:], y_hard[:] = np.nan, np.nan, np.nan, \
        np.nan

    # Fill Y vector
    y_sigma[idx_sigma_win - start_sf] = sigma_nfpow[idx_sigma_win]
    y_spindles[idx_spindles_win - start_sf] = data[idx_spindles_win]
    y_wlt[idx_spindles_win - start_sf] = amplitude[idx_spindles_win]
    y_hard[idx_hard_win - start_sf] = amplitude[idx_hard_win]

    # Start plot
    f, axarr = plt.subplots(4, figsize=(10, 6), sharex=True)
    f.subplots_adjust(hspace=0.6)

    # Plot original signal
    axarr[0].plot(x, data[x], 'darkslategrey', lw=1.5)
    axarr[0].plot(x, y_spindles, 'brown', lw=1.5)
    axarr[0].set_title('Original signal (' + str(window_s) + ' sec)')
    axarr[0].set_xlim([min(x), max(x)])

    if sp_power.size >= 1:
        text = 'power = ' + str(np.round(sp_power, 2)) + \
            ' - duration = ' + str(sp_duration) + ' ms'
        axarr[0].annotate(text, xy=(min(x), min(data[x])), fontsize=9,
                          xycoords='data')

    # Plot filtered signal
    axarr[1].plot(x, elec_filt, 'darkslategrey', linewidth=1.5)
    axarr[1].set_title('Filtered')

    # Plot wavelet envelope
    axarr[2].plot(x, amplitude[x], 'darkslategrey', linewidth=2)
    axarr[2].plot(x, y_wlt, 'coral', lw=3)
    # axarr[2].plot(x, y_hard, 'indianred', lw=3)
    axarr[2].scatter(idx_start_win, amplitude[idx_start_win], 60, 'coral')
    axarr[2].scatter(idx_stop_win, amplitude[idx_stop_win], 60, 'coral')
    axarr[2].set_title('Wavelet')
    axarr[2].axhline(y=hard_thr, color='grey', linestyle=':', lw=1.5)
    axarr[2].axhline(y=soft_thr, color='grey', linestyle=':', lw=1.5)

    # Plot sigma normalized power
    axarr[3].plot(x, sigma_nfpow[x], 'darkslategrey', linewidth=2)
    axarr[3].set_title('Sigma power')
    axarr[3].plot(x, y_sigma, lw=3, color='lightcoral')
    axarr[3].axhline(y=sigma_thr, color='grey', linestyle=':', lw=1.5)

    # Despine
    for ax in range(4):
        axarr[ax].axis('off')

    # Save as .png
    if isinstance(file, str):
        plt.savefig(file, format='png', dpi=dpi, bbox_inches='tight')
        logger.info('Image successfully saved (%s)' % file)
        plt.close()
    else:
        plt.show()


def write_fig_canvas(filename, canvas, widget=None, autocrop=False,
                     region=None, print_size=None, unit='centimeter', dpi=300.,
                     factor=1., bgcolor=None, transparent=False):
    """Export a canvas as a figure.

    Parameters
    ----------
    filename : string
        Name of the figure to export.
    canvas : VisPy canvas
        The vispy canvas to export.
    widget : PyQt widget | None
        The widget parent of the canvas.
    autocrop : bool | False
        Auto-cropping argument to remove useless space.
    region : tuple/list | None
        The region to export (x_start, y_start, width, height).
    print_size : tuple | None
        The desired print size. This argument should be used in association
        with the dpi and unit inputs. print_size describe should be a tuple
        of two floats describing (width, height) of the exported image for
        a specific dpi level. The final image might not have the exact
        desired size but will try instead to find a compromize
        regarding to the proportion of width/height of the original image.
    unit : {'centimeter', 'millimeter', 'pixel', 'inch'}
        Unit of the printed size.
    dpi : float | 300.
        Dots per inch for printing the image.
    factor : float | None
        If you don't want to use the print_size input, factor simply
        multiply the resolution of your screen.
    bgcolor : array_like/string | None
        Background color of the canvas.
    transparent : bool | False
        Use transparent background.
    """
    from ..utils import piccrop
    from vispy.io import imsave

    # Get the size of the canvas and backend :
    c_size = canvas.size
    b_size = canvas._backend._physical_size

    # If the GUI is displayed, c_size and b_size should be equals. If not,
    # and if the canvas is resizable, the canvas might have a different size
    # because it hasn't been updated. In that case, we force the canvas to have
    # the same size as the backend :
    if c_size != b_size:
        canvas.size = b_size

    # Backup size / background color :
    backup_size = canvas.physical_size
    backup_bgcolor = canvas.bgcolor

    # dpi checking :
    if print_size is None:
        logger.warning("dpi parameter is not active if `print_size` is None. "
                       "Use for example `print_size=(5, 5)`")

    # User select a desired print size with at a specific dpi :
    if print_size is not None:
        # Type checking :
        if not isinstance(print_size, (tuple, list)):
            raise TypeError("The print_size must either be a tuple or a list "
                            "describing the (width, height) of the"
                            " image in %s" % unit)
        # Check print size :
        if not all([isinstance(k, (int, float)) for k in print_size]):
            raise TypeError("print_size must be a tuple describing the "
                            "(width, height) of the image in %s" % unit)

        print_size = np.asarray(print_size)
        # If the user select the auto-croping option, the canvas must be render
        # before :
        if autocrop:
            img = canvas.render()
            s_output = piccrop(img)[:, :, 0].shape
            logger.info("Image cropped to closest non-backround pixels")
        else:
            s_output = b_size
        # Unit conversion :
        if unit == 'millimeter':
            mult = 1. / (10. * 2.54)
        elif unit == 'centimeter':
            mult = 1. / 2.54
        elif unit == 'pixel':
            mult = 1. / dpi
        elif unit == 'inch':
            mult = 1.
        else:
            raise ValueError("The unit must either be 'millimeter', "
                             "'centimeter', 'pixel' or 'inch' and not " + unit)
        # Get the factor to apply to the canvas size. This factor is defined as
        # the mean required float to get either the desired width/height.
        # Note that the min or the max can also be used instead.
        factor = np.mean(print_size * dpi * mult / np.asarray(s_output))

    # Multply the original canvas size :
    if factor is not None:
        # Get the new width and height :
        new_width = int(b_size[0] * factor)
        new_height = int(b_size[1] * factor)
        # Set it to the canvas, backend and the widget :
        canvas._backend._vispy_set_physical_size(new_width, new_height)
        canvas.size = (new_width, new_height)
        if widget is not None:
            widget.size = (new_width, new_height)

    # Background color and transparency :
    if bgcolor is not None:
        canvas.bgcolor = color2vb(bgcolor, alpha=1.)
    if transparent:
        canvas.bgcolor = [0.] * 4

    # Render the canvas :
    try:
        img = canvas.render(region=region)
    except:
        raise ValueError("Can not render the canvas. Try to decrease the "
                         "resolution")

    # Remove alpha for files that are not png or tiff :
    if os.path.splitext(filename)[1] not in ['.png', '.tiff']:
        img = img[..., 0:-1]

    # Apply auto-cropping to the image :
    if autocrop:
        img = piccrop(img)
        logger.info("Image cropped to closest non-backround pixels")
    # Save it :
    imsave(filename, img)
    px = tuple(img[:, :, 0].T.shape)
    logger.info("Image of size %rpx successfully saved (%s)" % (px, filename))

    # Set to the canvas it's previous size :
    canvas._backend._physical_size = backup_size
    canvas.size = backup_size
    canvas.bgcolor = backup_bgcolor


def write_fig_pyqt(self, filename):
    """Export a GUI window as a figure.

    Parameters
    ----------
    self : struct
        The self structure of the GUI window.
    filename : string
        The picture file name.
    """
    from PyQt5 import QtWidgets, QtCore

    # Screnshot function :
    def _take_screenshot():
        """Take the screenshot."""
        screen = QtWidgets.QApplication.primaryScreen()
        p = screen.grabWindow(0)
        p.save(filename)
        logger.info('Image successfully saved (%s)' % filename)
    # Take screenshot if filename :
    if filename:
        # Timer (avoid shooting the saving window)
        self.timerScreen = QtCore.QTimer()
        # self.timerScreen.setInterval(100)
        self.timerScreen.setSingleShot(True)
        self.timerScreen.timeout.connect(_take_screenshot)
        self.timerScreen.start(500)



"""Dialog window for saving and loading files.

* dialog_save : Open a window to save a file
* dialog_load : Open a window to load a file
"""
from PyQt5.QtWidgets import QFileDialog, QColorDialog
import os

from .rw_utils import safety_save

__all__ = ['dialog_save', 'dialog_load', 'dialog_color']


def dialog_save(self, name='Save file', default='file',
                allext=['All files (*.*)']):
    """Open a window to save a file.

    Parameters
    ----------
    self : class
        Class containing PyQt5 elemnets.
    name : string | 'Save file'
        Name of the saving window.
    default : string | 'file'
        Default name of the saved file.
    allext : list | ['All files (*.*)']
        String containing all the extensions. Must be a list where each
        element is a string of type 'Ext (.ext)'

    Returns
    -------
    filename : string
        Filename for saving.
    """
    # Build all extensions :
    if isinstance(allext, (list, tuple)):
        allext = ';;'.join(allext)
    # Open the window :
    file, ext = QFileDialog.getSaveFileName(self, name, default, allext)
    # By default, use the extension in the ruler :
    file = os.path.splitext(str(file))[0]
    ext = os.path.splitext(str(ext))[1][0:-1].lower()
    return safety_save(file + ext)


def dialog_load(self, name='Open file', default='file',
                allext=['All files (*.*)']):
    """Open a window to load a file.

    Parameters
    ----------
    self : class
        Class containing PyQt5 elemnets.
    name : string | 'Save file'
        Name of the opening window.
    default : string | 'file'
        Default name of the opened file.
    allext : list | ['All files (*.*)']
        String containing all the extensions. Must be a list where each
        element is a string of type 'Ext (.ext)'

    Returns
    -------
    filename : string
        Filename for opening.
    """
    # Open the window :
    file, _ = QFileDialog.getOpenFileName(self, name, default, allext)
    return str(file)


def dialog_color():
    """Open a QColorDialog window."""
    return QColorDialog.getColor().name()
